"""This module handles AnchorPy errors."""
from __future__ import annotations

import re
from enum import IntEnum
from typing import Dict, List, Optional, Tuple

from solders.pubkey import Pubkey
from solders.rpc.errors import SendTransactionPreflightFailureMessage
from solders.rpc.responses import RPCError
from solders.transaction_status import (
    InstructionErrorCustom,
    TransactionErrorInstructionError,
    TransactionErrorType,
)


class AccountDoesNotExistError(Exception):
    """Raise if account doesn't exist."""


class AccountInvalidDiscriminator(Exception):
    """Raise if account discriminator doesn't match the IDL."""


class IdlNotFoundError(Exception):
    """Raise when requested IDL account does not exist."""


class ArgsError(Exception):
    """Raise when the incorrect number of args is passed to the RPC function."""


class _LangErrorCode(IntEnum):
    """Enumerates Anchor error codes."""

    # Instructions.
    InstructionMissing = 100
    InstructionFallbackNotFound = 101
    InstructionDidNotDeserialize = 102
    InstructionDidNotSerialize = 103
    # IDL instructions.
    IdlInstructionStub = 1000
    IdlInstructionInvalidProgram = 1001
    # Constraints.
    ConstraintMut = 2000
    ConstraintHasOne = 2001
    ConstraintSigner = 2002
    ConstraintRaw = 2003
    ConstraintOwner = 2004
    ConstraintRentExempt = 2005
    ConstraintSeeds = 2006
    ConstraintExecutable = 2007
    ConstraintState = 2008
    ConstraintAssociated = 2009
    ConstraintAssociatedInit = 2010
    ConstraintClose = 2011
    ConstraintAddress = 2012
    ConstraintZero = 2013
    ConstraintTokenMint = 2014
    ConstraintTokenOwner = 2015
    ConstraintMintMintAuthority = 2016
    ConstraintMintFreezeAuthority = 2017
    ConstraintMintDecimals = 2018
    ConstraintSpace = 2019
    # Require
    RequireViolated = 2500
    RequireEqViolated = 2501
    RequireKeysEqViolated = 2502
    RequireNeqViolated = 2503
    RequireKeysNeqViolated = 2504
    RequireGtViolated = 2505
    RequireGteViolated = 2506
    # Accounts.
    AccountDiscriminatorAlreadySet = 3000
    AccountDiscriminatorNotFound = 3001
    AccountDiscriminatorMismatch = 3002
    AccountDidNotDeserialize = 3003
    AccountDidNotSerialize = 3004
    AccountNotEnoughKeys = 3005
    AccountNotMutable = 3006
    AccountOwnedByWrongProgram = 3007
    InvalidProgramId = 3008
    InvalidProgramExecutable = 3009
    AccountNotSigner = 3010
    AccountNotSystemOwned = 3011
    AccountNotInitialized = 3012
    AccountNotProgramData = 3013
    AccountNotAssociatedTokenAccount = 3014
    AccountSysvarMismatch = 3015
    # State.
    StateInvalidAddress = 4000

    # Used for APIs that shouldn't be used anymore.
    Deprecated = 5000


LangErrorMessage: Dict[int, str] = {
    # Instructions.
    _LangErrorCode.InstructionMissing: "8 byte instruction identifier not provided",
    _LangErrorCode.InstructionFallbackNotFound: "Fallback functions are not supported",
    _LangErrorCode.InstructionDidNotDeserialize: (
        "The program could not deserialize the given instruction"
    ),
    _LangErrorCode.InstructionDidNotSerialize: (
        "The program could not serialize the given instruction"
    ),
    # Idl instructions.
    _LangErrorCode.IdlInstructionStub: (
        "The program was compiled without idl instructions"
    ),
    _LangErrorCode.IdlInstructionInvalidProgram: (
        "The transaction was given an invalid program for the IDL instruction"
    ),
    # Constraints.
    _LangErrorCode.ConstraintMut: "A mut constraint was violated",
    _LangErrorCode.ConstraintHasOne: "A has_one constraint was violated",
    _LangErrorCode.ConstraintSigner: "A signer constraint was violated",
    _LangErrorCode.ConstraintRaw: "A raw constraint was violated",
    _LangErrorCode.ConstraintOwner: "An owner constraint was violated",
    _LangErrorCode.ConstraintRentExempt: "A rent exempt constraint was violated",
    _LangErrorCode.ConstraintSeeds: "A seeds constraint was violated",
    _LangErrorCode.ConstraintExecutable: "An executable constraint was violated",
    _LangErrorCode.ConstraintState: "A state constraint was violated",
    _LangErrorCode.ConstraintAssociated: "An associated constraint was violated",
    _LangErrorCode.ConstraintAssociatedInit: (
        "An associated init constraint was violated"
    ),
    _LangErrorCode.ConstraintClose: "A close constraint was violated",
    _LangErrorCode.ConstraintAddress: "An address constraint was violated",
    _LangErrorCode.ConstraintZero: "Expected zero account discriminant",
    _LangErrorCode.ConstraintTokenMint: "A token mint constraint was violated",
    _LangErrorCode.ConstraintTokenOwner: "A token owner constraint was violated",
    _LangErrorCode.ConstraintMintMintAuthority: (
        "A mint mint authority constraint was violated"
    ),
    _LangErrorCode.ConstraintMintFreezeAuthority: (
        "A mint freeze authority constraint was violated"
    ),
    _LangErrorCode.ConstraintMintDecimals: "A mint decimals constraint was violated",
    _LangErrorCode.ConstraintSpace: "A space constraint was violated",
    # Require.
    _LangErrorCode.RequireViolated: "A require expression was violated",
    _LangErrorCode.RequireEqViolated: "A require_eq expression was violated",
    _LangErrorCode.RequireKeysEqViolated: "A require_keys_eq expression was violated",
    _LangErrorCode.RequireNeqViolated: "A require_neq expression was violated",
    _LangErrorCode.RequireKeysNeqViolated: "A require_keys_neq expression was violated",
    _LangErrorCode.RequireGtViolated: "A require_gt expression was violated",
    _LangErrorCode.RequireGteViolated: "A require_gte expression was violated",
    # Accounts.
    _LangErrorCode.AccountDiscriminatorAlreadySet: (
        "The account discriminator was already set on this account"
    ),
    _LangErrorCode.AccountDiscriminatorNotFound: (
        "No 8 byte discriminator was found on the account"
    ),
    _LangErrorCode.AccountDiscriminatorMismatch: (
        "8 byte discriminator did not match what was expected"
    ),
    _LangErrorCode.AccountDidNotDeserialize: "Failed to deserialize the account",
    _LangErrorCode.AccountDidNotSerialize: "Failed to serialize the account",
    _LangErrorCode.AccountNotEnoughKeys: (
        "Not enough account keys given to the instruction"
    ),
    _LangErrorCode.AccountNotMutable: "The given account is not mutable",
    _LangErrorCode.AccountOwnedByWrongProgram: (
        "The given account is owned by a different program than expected"
    ),
    _LangErrorCode.InvalidProgramId: "Program ID was not as expected",
    _LangErrorCode.InvalidProgramExecutable: "Program account is not executable",
    _LangErrorCode.AccountNotSigner: "The given account did not sign",
    _LangErrorCode.AccountNotSystemOwned: (
        "The given account is not owned by the system program"
    ),
    _LangErrorCode.AccountNotInitialized: (
        "The program expected this account to be already initialized"
    ),
    _LangErrorCode.AccountNotProgramData: (
        "The given account is not a program data account"
    ),
    _LangErrorCode.AccountNotAssociatedTokenAccount: (
        "The given account is not the associated token account"
    ),
    _LangErrorCode.AccountSysvarMismatch: (
        "The given public key does not match the required sysvar"
    ),
    # State.
    _LangErrorCode.StateInvalidAddress: (
        "The given state account does not have the correct address"
    ),
    # Misc.
    _LangErrorCode.Deprecated: (
        "The API being used is deprecated and should no longer be used"
    ),
}


class ProgramError(Exception):
    """An error from a user defined program."""

    def __init__(
        self, code: int, msg: Optional[str], logs: Optional[list[str]] = None
    ) -> None:
        """Init.

        Args:
            code: The error code.
            msg: The error message.
            logs: The transaction simulation logs.
        """
        self.code = code
        self.msg = msg
        self.logs = logs
        super().__init__(f"{code}: {msg}")

    @classmethod
    def parse(
        cls,
        err_info: RPCError,
        idl_errors: dict[int, str],
        program_id: Pubkey,
    ) -> Optional[ProgramError]:
        """Convert an RPC error into a ProgramError, if possible.

        Args:
            err_info: The RPC error.
            idl_errors: Errors from the IDL file.
            program_id: The ID of the program we expect the error to come from.

        Returns:
            A ProgramError or None.
        """
        extracted = extract_code_and_logs(err_info, program_id)
        if extracted is None:
            return None
        code, logs = extracted
        msg = idl_errors.get(code)
        if msg is not None:
            return cls(code, msg, logs)
        # parse framework internal error
        msg = LangErrorMessage.get(code)
        if msg is not None:
            return cls(code, msg, logs)
        # Unable to parse the error.
        return None

    @classmethod
    def parse_tx_error(
        cls,
        err_info: TransactionErrorType,
        idl_errors: dict[int, str],
        program_id: Pubkey,
        logs: List[str],
    ) -> Optional[ProgramError]:
        """Convert an RPC error into a ProgramError, if possible.

        Args:
            err_info: The RPC error.
            idl_errors: Errors from the IDL file.
            program_id: The ID of the program we expect the error to come from.
            logs: The transaction logs.

        Returns:
            A ProgramError or None.
        """
        code = extract_code_tx_error(err_info, program_id, logs)
        if code is None:
            return None
        msg = idl_errors.get(code)
        if msg is not None:
            return cls(code, msg, logs)
        # parse framework internal error
        msg = LangErrorMessage.get(code)
        if msg is not None:
            return cls(code, msg, logs)
        # Unable to parse the error.
        return None


error_re = re.compile(r"Program (\w+) failed: custom program error: (\w+)")


def _find_first_match(logs: list[str]) -> Optional[re.Match]:
    for logline in logs:
        first_match = error_re.match(logline)
        if first_match is not None:
            return first_match
    return None


def extract_code_and_logs(
    err_info: RPCError, program_id: Pubkey
) -> Optional[Tuple[int, List[str]]]:
    """Extract the custom instruction error code from an RPC response.

    Args:
        err_info: The RPC error.
        program_id: The ID of the program we expect the error to come from.
    """
    if isinstance(err_info, SendTransactionPreflightFailureMessage):
        err_data = err_info.data
        err_data_err = err_data.err
        logs = err_data.logs
        if logs is None:
            return None
        if err_data_err is None:
            return None
        maybe_code = _handle_ix_err(err_data_err, logs, program_id)
        return None if maybe_code is None else (maybe_code, logs)
    return None


def extract_code_tx_error(
    err_info: TransactionErrorType, program_id: Pubkey, logs: List[str]
) -> Optional[int]:
    """Extract the custom instruction error code from a transaction error.

    Args:
        err_info: The tx error.
        program_id: The ID of the program we expect the error to come from.
        logs: The tx logs.
    """
    return _handle_ix_err(err_info, logs, program_id)


def _handle_ix_err(
    err: TransactionErrorType, logs: List[str], program_id: Pubkey
) -> Optional[int]:
    if isinstance(err, TransactionErrorInstructionError):
        instruction_err = err.err
        if isinstance(instruction_err, InstructionErrorCustom):
            code = instruction_err.code
            first_match = _find_first_match(logs)
            if first_match is None:
                return None
            program_id_raw, _ = first_match.groups()
            if program_id_raw != str(program_id):
                return None
            return code
    return None



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class TypeAliasArgs(typing.TypedDict):
    u8_array: types.u8_array.U8Array


layout = borsh.CStruct("u8_array" / borsh.U8[8])


def type_alias(
    args: TypeAliasArgs,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = []
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"5\x84\x11\xf9y\xc216"
    encoded_args = layout.build(
        {
            "u8_array": args["u8_array"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



#!/usr/bin/env python3
"""Test minimal IDL structures to isolate the issue."""

import json
from anchorpy3_core.idl import Idl

# Test 1: Empty IDL with null version
print("Test 1: Empty IDL with null version")
minimal = {
    "version": None,
    "name": "test",
    "instructions": [],
    "accounts": [],
    "types": [],
    "errors": []
}

try:
    idl = Idl.from_json(json.dumps(minimal))
    print("  ✅ Parsed successfully")
    print(f"     Version: {idl.version}")
    print(f"     Name: {idl.name}")
except Exception as e:
    print(f"  ❌ Failed: {e}")

# Test 2: Add metadata
print("\nTest 2: With metadata")
minimal["metadata"] = {
    "name": "test",
    "version": "0.1.0",
    "spec": "0.1.0"
}

try:
    idl = Idl.from_json(json.dumps(minimal))
    print("  ✅ Parsed successfully")
except Exception as e:
    print(f"  ❌ Failed: {e}")

# Test 3: Add a new format account
print("\nTest 3: With new format account")
minimal["accounts"] = [
    {
        "name": "TestAccount",
        "discriminator": [1, 2, 3, 4, 5, 6, 7, 8]
    }
]

try:
    idl = Idl.from_json(json.dumps(minimal))
    print("  ✅ Parsed successfully")
    print(f"     Accounts: {len(idl.accounts)}")
except Exception as e:
    print(f"  ❌ Failed: {e}")

# Test 4: Add an error
print("\nTest 4: With error")
minimal["errors"] = [
    {
        "code": 6000,
        "name": "TestError",
        "msg": "Test error message"
    }
]

try:
    idl = Idl.from_json(json.dumps(minimal))
    print("  ✅ Parsed successfully")
    print(f"     Errors: {len(idl.errors) if idl.errors else 0}")
except Exception as e:
    print(f"  ❌ Failed: {e}")

# Test 5: Load actual loopscale_v2.json errors section
print("\nTest 5: With loopscale_v2.json errors")
with open('../loopscale_v2.json') as f:
    v2_data = json.load(f)

minimal["errors"] = v2_data["errors"]

try:
    idl = Idl.from_json(json.dumps(minimal))
    print("  ✅ Parsed successfully")
    print(f"     Errors: {len(idl.errors) if idl.errors else 0}")
except Exception as e:
    print(f"  ❌ Failed: {e}")
    print(f"     Error position suggests issue around: {str(e)}")

# Test 6: Try with types
print("\nTest 6: With loopscale_v2.json types")
minimal["errors"] = []
minimal["types"] = v2_data["types"][:5]  # Try with first 5 types

try:
    idl = Idl.from_json(json.dumps(minimal))
    print("  ✅ Parsed successfully")
    print(f"     Types: {len(idl.types)}")
except Exception as e:
    print(f"  ❌ Failed: {e}")


"""Code generation utilities."""
import keyword
from typing import Optional

from anchorpy3_core.idl import (
    Idl,
    IdlField,
    IdlType,
    IdlTypeArray,
    IdlTypeDefined,
    IdlTypeDefinitionTyAlias,
    IdlTypeDefinitionTyStruct,
    IdlTypeOption,
    IdlTypeSimple,
    IdlTypeVec,
)
from pyheck import snake

from anchorpy.coder.idl_compat import get_defined_type_name

_DEFAULT_DEFINED_TYPES_PREFIX = "types."

INT_TYPES = {
    IdlTypeSimple.U8,
    IdlTypeSimple.I8,
    IdlTypeSimple.U16,
    IdlTypeSimple.I16,
    IdlTypeSimple.U32,
    IdlTypeSimple.I32,
    IdlTypeSimple.U64,
    IdlTypeSimple.I64,
    IdlTypeSimple.U128,
    IdlTypeSimple.I128,
}
FLOAT_TYPES = {IdlTypeSimple.F32, IdlTypeSimple.F64}
NUMBER_TYPES = INT_TYPES | FLOAT_TYPES


def _fields_interface_name(type_name: str) -> str:
    return f"{type_name}Fields"


def _value_interface_name(type_name: str) -> str:
    return f"{type_name}Value"


def _kind_interface_name(type_name: str) -> str:
    return f"{type_name}Kind"


def _json_interface_name(type_name: str) -> str:
    return f"{type_name}JSON"


def _sanitize(name: str) -> str:
    # Replace :: with _ for Rust-style paths
    name = name.replace('::', '_')
    # Add underscore if it's a Python keyword
    return f"{name}_" if keyword.iskeyword(name) else name


def _py_type_from_idl(
    idl: Idl,
    ty: IdlType,
    types_relative_imports: bool,
    use_fields_interface_for_struct: bool,
) -> str:
    if isinstance(ty, IdlTypeVec):
        inner_type = _py_type_from_idl(
            idl=idl,
            ty=ty.vec,
            types_relative_imports=types_relative_imports,
            use_fields_interface_for_struct=use_fields_interface_for_struct,
        )
        return f"list[{inner_type}]"
    if isinstance(ty, IdlTypeOption):
        inner_type = _py_type_from_idl(
            idl=idl,
            ty=ty.option,
            types_relative_imports=types_relative_imports,
            use_fields_interface_for_struct=use_fields_interface_for_struct,
        )
        return f"typing.Optional[{inner_type}]"
    if isinstance(ty, IdlTypeDefined):
        # Support both old string format and new object format
        type_name = get_defined_type_name(ty.defined)
        defined = _sanitize(type_name)
        filtered = [t for t in idl.types if _sanitize(t.name) == defined]
        maybe_coption_split = defined.split("COption<")
        if len(maybe_coption_split) == 2:
            inner_type = {"u64": "int", "Pubkey": "Pubkey"}[maybe_coption_split[1][:-1]]
            return f"typing.Optional[{inner_type}]"
        if defined == "&'astr":
            return "str"
        defined_types_prefix = (
            "" if types_relative_imports else _DEFAULT_DEFINED_TYPES_PREFIX
        )
        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        typedef_type = filtered[0].ty
        module = _sanitize(snake(type_name))
        if isinstance(typedef_type, IdlTypeDefinitionTyStruct):
            name = (
                _fields_interface_name(defined)  # Use sanitized name
                if use_fields_interface_for_struct
                else defined
            )
        elif isinstance(typedef_type, IdlTypeDefinitionTyAlias):
            name = defined
        else:
            # enum
            name = _kind_interface_name(defined)  # Use sanitized name
        return f"{defined_types_prefix}{module}.{name}"
    if isinstance(ty, IdlTypeArray):
        inner_type = _py_type_from_idl(
            idl=idl,
            ty=ty.array[0],
            types_relative_imports=types_relative_imports,
            use_fields_interface_for_struct=use_fields_interface_for_struct,
        )
        return f"list[{inner_type}]"
    if ty in {IdlTypeSimple.Bool, IdlTypeSimple.Bytes}:
        return str(ty).replace("IdlTypeSimple.", "").lower()
    if ty in INT_TYPES:
        return "int"
    if ty in FLOAT_TYPES:
        return "float"
    if ty == IdlTypeSimple.String:
        return "str"
    if ty == IdlTypeSimple.PublicKey:
        return "Pubkey"
    raise ValueError(f"Unrecognized type: {ty}")


def _layout_for_type(
    idl: Idl,
    ty: IdlType,
    types_relative_imports: bool,
    name: Optional[str] = None,
) -> str:
    if ty == IdlTypeSimple.PublicKey:
        inner = "BorshPubkey"
    elif isinstance(ty, IdlTypeSimple):
        inner = str(ty).replace("IdlTypeSimple", "borsh")
    elif isinstance(ty, IdlTypeVec):
        layout = _layout_for_type(
            idl=idl, ty=ty.vec, types_relative_imports=types_relative_imports
        )
        cast_layout = f"typing.cast(Construct, {layout})"
        inner = f"borsh.Vec({cast_layout})"
    elif isinstance(ty, IdlTypeOption):
        layout = _layout_for_type(
            idl=idl, ty=ty.option, types_relative_imports=types_relative_imports
        )
        inner = f"borsh.Option({layout})"
    elif isinstance(ty, IdlTypeDefined):
        # Support both old string format and new object format
        type_name = get_defined_type_name(ty.defined)
        defined = _sanitize(type_name)
        maybe_coption_split = defined.split("COption<")
        if len(maybe_coption_split) == 2:
            layout_str = maybe_coption_split[1][:-1]
            layout = {"u64": "borsh.U64", "Pubkey": "BorshPubkey"}[layout_str]
            inner = f"COption({layout})"
        elif defined == "&'astr":
            return "borsh.String"
        else:
            filtered = [t for t in idl.types if _sanitize(t.name) == defined]
            typedef_type = filtered[0].ty
            if isinstance(typedef_type, IdlTypeDefinitionTyAlias):
                return _layout_for_type(
                    idl, typedef_type.value, types_relative_imports=True, name=name
                )
            defined_types_prefix = (
                "" if types_relative_imports else _DEFAULT_DEFINED_TYPES_PREFIX
            )
            module = snake(defined)
            if isinstance(typedef_type, IdlTypeDefinitionTyStruct):
                inner = f"{defined_types_prefix}{module}.{defined}.layout"
            else:
                inner = f"{defined_types_prefix}{module}.layout"
    elif isinstance(ty, IdlTypeArray):
        layout = _layout_for_type(
            idl=idl, ty=ty.array[0], types_relative_imports=types_relative_imports
        )
        inner = f"{layout}[{ty.array[1]}]"
    else:
        raise ValueError(f"Unrecognized type: {ty}")

    if name is None:
        return inner
    return f'"{name}" / {inner}'


def _maybe_none(to_check: str, if_not_none: str) -> str:
    return f"(None if {to_check} is None else {if_not_none})"


def _field_to_encodable(
    idl: Idl,
    ty: IdlField,
    types_relative_imports: bool,
    val_prefix: str = "",
    val_suffix: str = "",
    convert_case: bool = True,
) -> str:
    ty_type = ty.ty
    maybe_converted = snake(ty.name) if convert_case else ty.name
    ty_name = _sanitize(maybe_converted)
    if isinstance(ty_type, IdlTypeVec):
        map_body = _field_to_encodable(
            idl=idl,
            ty=IdlField("item", docs=None, ty=ty_type.vec),
            val_prefix="",
            types_relative_imports=types_relative_imports,
            val_suffix="",
        )
        # skip mapping when not needed
        if map_body == "item":
            return f"{val_prefix}{ty_name}{val_suffix}"
        return f"list(map(lambda item: {map_body}, {val_prefix}{ty_name}{val_suffix}))"
    if isinstance(ty_type, IdlTypeOption):
        encodable = _field_to_encodable(
            idl=idl,
            ty=IdlField(ty_name, docs=None, ty=ty_type.option),
            val_prefix=val_prefix,
            types_relative_imports=types_relative_imports,
            val_suffix=val_suffix,
            convert_case=convert_case,
        )
        if encodable == f"{val_prefix}{ty_name}{val_suffix}":
            return encodable
        return _maybe_none(f"{val_prefix}{ty_name}{val_suffix}", encodable)
    if isinstance(ty_type, IdlTypeDefined):
        defined = _sanitize(ty_type.defined)
        maybe_coption_split = defined.split("COption<")
        if len(maybe_coption_split) == 2:
            inner_type = maybe_coption_split[1][:-1]
            encodable = _field_to_encodable(
                idl=idl,
                ty=IdlField(
                    ty_name,
                    docs=None,
                    ty={"u64": IdlTypeSimple.U64, "Pubkey": IdlTypeSimple.PublicKey}[
                        inner_type
                    ],
                ),
                val_prefix=val_prefix,
                types_relative_imports=types_relative_imports,
                val_suffix=val_suffix,
                convert_case=convert_case,
            )
            if encodable == f"{val_prefix}{ty_name}{val_suffix}":
                return encodable
            return _maybe_none(f"{val_prefix}{ty_name}{val_suffix}", encodable)
        elif defined == "&'astr":
            return f"{val_prefix}{ty_name}{val_suffix}"
        filtered = [t for t in idl.types if _sanitize(t.name) == defined]

        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        typedef_type = filtered[0].ty
        if isinstance(typedef_type, IdlTypeDefinitionTyStruct):
            val_full_name = f"{val_prefix}{ty_name}{val_suffix}"
            return f"{val_full_name}.to_encodable()"
        if isinstance(typedef_type, IdlTypeDefinitionTyAlias):
            return f"{val_prefix}{ty_name}{val_suffix}"
        return f"{val_prefix}{ty_name}{val_suffix}.to_encodable()"
    if isinstance(ty_type, IdlTypeArray):
        map_body = _field_to_encodable(
            idl=idl,
            ty=IdlField("item", docs=None, ty=ty_type.array[0]),
            val_prefix="",
            types_relative_imports=types_relative_imports,
            val_suffix="",
        )
        # skip mapping when not needed
        if map_body == "item":
            return f"{val_prefix}{ty_name}{val_suffix}"
        return f"list(map(lambda item: {map_body}, {val_prefix}{ty_name}{val_suffix}))"
    if ty_type in {
        IdlTypeSimple.Bool,
        *NUMBER_TYPES,
        IdlTypeSimple.String,
        IdlTypeSimple.PublicKey,
        IdlTypeSimple.Bytes,
    }:
        return f"{val_prefix}{ty_name}{val_suffix}"
    raise ValueError(f"Unrecognized type: {ty_type}")


def _field_from_decoded(
    idl: Idl, ty: IdlField, types_relative_imports: bool, val_prefix: str = ""
) -> str:
    ty_type = ty.ty
    ty_name = _sanitize(ty.name)
    if isinstance(ty_type, IdlTypeVec):
        map_body = _field_from_decoded(
            idl=idl,
            ty=IdlField("item", docs=None, ty=ty_type.vec),
            val_prefix="",
            types_relative_imports=types_relative_imports,
        )
        # skip mapping when not needed
        if map_body == "item":
            return f"{val_prefix}{ty_name}"
        return f"list(map(lambda item: {map_body}, {val_prefix}{ty_name}))"
    if isinstance(ty_type, IdlTypeOption):
        decoded = _field_from_decoded(
            idl=idl,
            ty=IdlField(ty_name, docs=None, ty=ty_type.option),
            types_relative_imports=types_relative_imports,
            val_prefix=val_prefix,
        )
        # skip coercion when not needed
        if decoded == f"{val_prefix}{ty_name}":
            return decoded
        return _maybe_none(f"{val_prefix}{ty_name}", decoded)
    if isinstance(ty_type, IdlTypeDefined):
        defined = _sanitize(ty_type.defined)
        maybe_coption_split = defined.split("COption<")
        if len(maybe_coption_split) == 2:
            inner_type = maybe_coption_split[1][:-1]
            decoded = _field_from_decoded(
                idl=idl,
                ty=IdlField(
                    ty_name,
                    docs=None,
                    ty={"u64": IdlTypeSimple.U64, "Pubkey": IdlTypeSimple.PublicKey}[
                        inner_type
                    ],
                ),
                types_relative_imports=types_relative_imports,
                val_prefix=val_prefix,
            )
            # skip coercion when not needed
            if decoded == f"{val_prefix}{ty_name}":
                return decoded
            return _maybe_none(f"{val_prefix}{ty_name}", decoded)
        filtered = [t for t in idl.types if _sanitize(t.name) == defined]
        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        typedef_type = filtered[0].ty
        from_decoded_func_path = (
            f"{snake(defined)}.{defined}"
            if isinstance(typedef_type, IdlTypeDefinitionTyStruct)
            else f"{snake(defined)}"
        )
        defined_types_prefix = (
            "" if types_relative_imports else _DEFAULT_DEFINED_TYPES_PREFIX
        )
        full_func_path = f"{defined_types_prefix}{from_decoded_func_path}"
        from_decoded_arg = f"{val_prefix}{ty_name}"
        return f"{full_func_path}.from_decoded({from_decoded_arg})"
    if isinstance(ty_type, IdlTypeArray):
        map_body = _field_from_decoded(
            idl=idl,
            ty=IdlField("item", docs=None, ty=ty_type.array[0]),
            val_prefix="",
            types_relative_imports=types_relative_imports,
        )
        # skip mapping when not needed
        if map_body == "item":
            return f"{val_prefix}{ty_name}"
        return f"list(map(lambda item: {map_body}, {val_prefix}{ty_name}))"
    if ty_type in {
        IdlTypeSimple.Bool,
        *NUMBER_TYPES,
        IdlTypeSimple.String,
        IdlTypeSimple.PublicKey,
        IdlTypeSimple.Bytes,
    }:
        return f"{val_prefix}{ty_name}"
    raise ValueError(f"Unrecognized type: {ty_type}")


def _struct_field_initializer(
    idl: Idl,
    field: IdlField,
    types_relative_imports: bool,
    prefix: str = 'fields["',
    suffix: str = '"]',
) -> str:
    field_type = field.ty
    field_name = _sanitize(snake(field.name))
    if isinstance(field_type, IdlTypeDefined):
        defined = _sanitize(field_type.defined)
        filtered = [t for t in idl.types if _sanitize(t.name) == defined]
        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        typedef_type = filtered[0].ty
        if isinstance(typedef_type, IdlTypeDefinitionTyStruct):
            module = snake(defined)
            defined_types_prefix = (
                "" if types_relative_imports else _DEFAULT_DEFINED_TYPES_PREFIX
            )
            obj_name = f"{defined_types_prefix}{module}.{defined}"
            return f"{obj_name}(**{prefix}{field_name}{suffix})"
        return f"{prefix}{field_name}{suffix}"
    if isinstance(field_type, IdlTypeOption):
        initializer = _struct_field_initializer(
            idl=idl,
            field=IdlField(field_name, docs=None, ty=field_type.option),
            prefix=prefix,
            suffix=suffix,
            types_relative_imports=types_relative_imports,
        )
        # skip coercion when not needed
        if initializer == f"{prefix}{field_name}{suffix}":
            return initializer
        return _maybe_none(f"{prefix}{field_name}{suffix}", initializer)
    if isinstance(field_type, IdlTypeArray):
        map_body = _struct_field_initializer(
            idl=idl,
            field=IdlField("item", docs=None, ty=field_type.array[0]),
            prefix="",
            suffix="",
            types_relative_imports=types_relative_imports,
        )
        # skip mapping when not needed
        if map_body == "item":
            return f"{prefix}{field_name}{suffix}"
        return f"list(map(lambda item: {map_body}, {prefix}{field_name}{suffix}))"
    if isinstance(field_type, IdlTypeVec):
        map_body = _struct_field_initializer(
            idl=idl,
            field=IdlField("item", docs=None, ty=field_type.vec),
            prefix="",
            suffix="",
            types_relative_imports=types_relative_imports,
        )
        # skip mapping when not needed
        if map_body == "item":
            return f"{prefix}{field_name}{suffix}"
        return f"list(map(lambda item: {map_body}, {prefix}{field_name}{suffix}))"
    if field_type in {
        IdlTypeSimple.Bool,
        *NUMBER_TYPES,
        IdlTypeSimple.String,
        IdlTypeSimple.PublicKey,
        IdlTypeSimple.Bytes,
    }:
        return f"{prefix}{field_name}{suffix}"
    raise ValueError(f"Unrecognized type: {field_type}")


def _field_to_json(
    idl: Idl,
    ty: IdlField,
    val_prefix: str = "",
    val_suffix: str = "",
    convert_case: bool = True,
) -> str:
    ty_type = ty.ty
    maybe_converted = _sanitize(snake(ty.name)) if convert_case else _sanitize(ty.name)
    var_name = f"{val_prefix}{maybe_converted}{val_suffix}"
    if ty_type == IdlTypeSimple.PublicKey:
        return f"str({var_name})"
    if isinstance(ty_type, IdlTypeVec):
        map_body = _field_to_json(idl, IdlField("item", docs=None, ty=ty_type.vec))
        # skip mapping when not needed
        if map_body == "item":
            return var_name
        return f"list(map(lambda item: {map_body}, {var_name}))"
    if isinstance(ty_type, IdlTypeArray):
        map_body = _field_to_json(idl, IdlField("item", docs=None, ty=ty_type.array[0]))
        # skip mapping when not needed
        if map_body == "item":
            return var_name
        return f"list(map(lambda item: {map_body}, {var_name}))"
    if isinstance(ty_type, IdlTypeOption):
        value = _field_to_json(
            idl,
            IdlField(ty.name, docs=None, ty=ty_type.option),
            val_prefix,
            val_suffix,
            convert_case=convert_case,
        )
        # skip coercion when not needed
        if value == var_name:
            return value
        return _maybe_none(var_name, value)
    if isinstance(ty_type, IdlTypeDefined):
        defined = ty_type.defined
        filtered = [t for t in idl.types if t.name == defined]
        maybe_coption_split = defined.split("COption<")
        if len(maybe_coption_split) == 2:
            inner_type = maybe_coption_split[1][:-1]
            value = _field_to_json(
                idl,
                IdlField(
                    ty.name,
                    docs=None,
                    ty={"u64": IdlTypeSimple.U64, "Pubkey": IdlTypeSimple.PublicKey}[
                        inner_type
                    ],
                ),
                val_prefix,
                val_suffix,
                convert_case=convert_case,
            )
            # skip coercion when not needed
            if value == var_name:
                return value
            return _maybe_none(var_name, value)
        if len(filtered) != 1:
            raise ValueError(f"Type not found {defined}")
        return f"{var_name}.to_json()"
    if ty_type == IdlTypeSimple.Bytes:
        return f"list({var_name})"
    if ty_type in {
        IdlTypeSimple.Bool,
        *NUMBER_TYPES,
        IdlTypeSimple.String,
    }:
        return var_name
    raise ValueError(f"Unrecognized type: {ty_type}")


def _idl_type_to_json_type(ty: IdlType, types_relative_imports: bool) -> str:
    if isinstance(ty, IdlTypeVec):
        inner = _idl_type_to_json_type(
            ty=ty.vec, types_relative_imports=types_relative_imports
        )
        return f"list[{inner}]"
    if isinstance(ty, IdlTypeArray):
        inner = _idl_type_to_json_type(
            ty=ty.array[0], types_relative_imports=types_relative_imports
        )
        return f"list[{inner}]"
    if isinstance(ty, IdlTypeOption):
        inner = _idl_type_to_json_type(
            ty=ty.option, types_relative_imports=types_relative_imports
        )
        return f"typing.Optional[{inner}]"
    if isinstance(ty, IdlTypeDefined):
        # Support both old string format and new object format
        type_name = get_defined_type_name(ty.defined)
        defined = _sanitize(type_name)  # Sanitize the name
        maybe_coption_split = type_name.split("COption<")  # Use original for COption check
        if len(maybe_coption_split) == 2:
            inner_type = {"u64": "int", "Pubkey": "str"}[maybe_coption_split[1][:-1]]
            return f"typing.Optional[{inner_type}]"
        defined_types_prefix = (
            "" if types_relative_imports else _DEFAULT_DEFINED_TYPES_PREFIX
        )
        module = _sanitize(snake(type_name))
        return f"{defined_types_prefix}{module}.{_json_interface_name(defined)}"
    if ty == IdlTypeSimple.Bool:
        return "bool"
    if ty in INT_TYPES:
        return "int"
    if ty in FLOAT_TYPES:
        return "float"
    if ty == IdlTypeSimple.Bytes:
        return "list[int]"
    if ty in {IdlTypeSimple.String, IdlTypeSimple.PublicKey}:
        return "str"
    raise ValueError(f"Unrecognized type: {ty}")


def _field_from_json(
    idl: Idl,
    ty: IdlField,
    types_relative_imports: bool,
    param_prefix: str = 'obj["',
    param_suffix: str = '"]',
) -> str:
    ty_type = ty.ty
    ty_name_snake_unsanitized = snake(ty.name)
    ty_name = _sanitize(ty_name_snake_unsanitized)
    var_name = f"{param_prefix}{ty_name_snake_unsanitized}{param_suffix}"
    if ty_type == IdlTypeSimple.PublicKey:
        return f"Pubkey.from_string({var_name})"
    if isinstance(ty_type, IdlTypeVec):
        map_body = _field_from_json(
            idl=idl,
            ty=IdlField("item", docs=None, ty=ty_type.vec),
            param_prefix="",
            param_suffix="",
            types_relative_imports=types_relative_imports,
        )
        # skip mapping when not needed
        if map_body == "item":
            return var_name
        return f"list(map(lambda item: {map_body}, {var_name}))"
    if isinstance(ty_type, IdlTypeArray):
        map_body = _field_from_json(
            idl=idl,
            ty=IdlField("item", docs=None, ty=ty_type.array[0]),
            param_prefix="",
            param_suffix="",
            types_relative_imports=types_relative_imports,
        )
        # skip mapping when not needed
        if map_body == "item":
            return var_name
        return f"list(map(lambda item: {map_body}, {var_name}))"
    if isinstance(ty_type, IdlTypeOption):
        inner = _field_from_json(
            idl=idl,
            ty=IdlField(ty_name, docs=None, ty=ty_type.option),
            param_prefix=param_prefix,
            param_suffix=param_suffix,
            types_relative_imports=types_relative_imports,
        )
        # skip coercion when not needed
        if inner == var_name:
            return inner
        return _maybe_none(var_name, inner)
    if isinstance(ty_type, IdlTypeDefined):
        from_json_arg = var_name
        defined = _sanitize(ty_type.defined)
        maybe_coption_split = defined.split("COption<")
        if len(maybe_coption_split) == 2:
            inner_type = maybe_coption_split[1][:-1]
            inner = _field_from_json(
                idl=idl,
                ty=IdlField(
                    ty_name,
                    docs=None,
                    ty={"u64": IdlTypeSimple.U64, "Pubkey": IdlTypeSimple.PublicKey}[
                        inner_type
                    ],
                ),
                param_prefix=param_prefix,
                param_suffix=param_suffix,
                types_relative_imports=types_relative_imports,
            )
            # skip coercion when not needed
            if inner == var_name:
                return inner
            return _maybe_none(var_name, inner)
        defined_snake = _sanitize(snake(ty_type.defined))
        filtered = [t for t in idl.types if _sanitize(t.name) == defined]
        typedef_type = filtered[0].ty
        from_json_func_path = (
            f"{defined_snake}.{defined}"
            if isinstance(typedef_type, IdlTypeDefinitionTyStruct)
            else f"{defined_snake}"
        )
        defined_types_prefix = (
            "" if types_relative_imports else _DEFAULT_DEFINED_TYPES_PREFIX
        )
        full_func_path = f"{defined_types_prefix}{from_json_func_path}"
        return f"{full_func_path}.from_json({from_json_arg})"
    if ty_type == IdlTypeSimple.Bytes:
        return f"bytes({var_name})"
    if ty_type in {
        IdlTypeSimple.Bool,
        *NUMBER_TYPES,
        IdlTypeSimple.String,
    }:
        return var_name
    raise ValueError(f"Unrecognized type: {ty_type}")



"""Test IDL compatibility for both old and new formats."""
import json
from pathlib import Path

import pytest
from anchorpy3_core.idl import Idl
from solders.pubkey import Pubkey

from anchorpy.coder.accounts import AccountsCoder
from anchorpy.coder.idl_compat import (
    detect_idl_format,
    get_account_discriminator,
    get_instruction_discriminator,
    get_defined_type_name,
    get_account_writable,
    get_account_signer,
)
from anchorpy.coder.instruction import InstructionCoder
from anchorpy.coder.event import EventCoder
from anchorpy import Program, Provider


class TestIDLFormatDetection:
    """Test IDL format detection."""

    def test_detect_old_format_kamino(self):
        """Test detection of old format IDL (kamino_lend_v4.json)."""
        idl_path = Path(__file__).parent.parent.parent / "kamino_lend_v4.json"
        if idl_path.exists():
            with open(idl_path) as f:
                idl_data = json.load(f)
                idl = Idl.from_json(json.dumps(idl_data))
                assert detect_idl_format(idl) == "old"

    def test_detect_new_format_loopscale(self):
        """Test detection of new format IDL (loopscale_v1.json)."""
        idl_path = Path(__file__).parent.parent.parent / "loopscale_v1.json"
        if idl_path.exists():
            with open(idl_path) as f:
                idl_data = json.load(f)
                # Check the raw data to determine format
                if "address" in idl_data and "metadata" in idl_data:
                    if "spec" in idl_data.get("metadata", {}):
                        # This is new format
                        assert True  # New format detected correctly

    def test_detect_old_format_basic_idl(self):
        """Test detection of old format IDL from test fixtures."""
        idl_path = Path(__file__).parent / "idls" / "basic_1.json"
        if idl_path.exists():
            with open(idl_path) as f:
                idl_data = json.load(f)
                idl = Idl.from_json(json.dumps(idl_data))
                assert detect_idl_format(idl) == "old"


class TestDiscriminatorHandling:
    """Test discriminator handling for both formats."""

    def test_account_discriminator_old_format(self):
        """Test that old format calculates discriminator correctly."""
        # Create a mock account object without discriminator field
        class MockAccount:
            def __init__(self):
                self.name = "TestAccount"

        acc = MockAccount()
        # Should return None since no discriminator field exists
        assert get_account_discriminator(acc) is None

    def test_account_discriminator_new_format(self):
        """Test that new format reads precomputed discriminator."""
        # Create a mock account object with discriminator field
        class MockAccount:
            def __init__(self):
                self.name = "TestAccount"
                self.discriminator = [1, 2, 3, 4, 5, 6, 7, 8]

        acc = MockAccount()
        # Should return the discriminator list
        assert get_account_discriminator(acc) == [1, 2, 3, 4, 5, 6, 7, 8]

    def test_instruction_discriminator_old_format(self):
        """Test that old format returns None for discriminator."""
        # Create a mock instruction object without discriminator field
        class MockInstruction:
            def __init__(self):
                self.name = "testInstruction"

        ix = MockInstruction()
        # Should return None since no discriminator field exists
        assert get_instruction_discriminator(ix) is None

    def test_instruction_discriminator_new_format(self):
        """Test that new format reads precomputed discriminator."""
        # Create a mock instruction object with discriminator field
        class MockInstruction:
            def __init__(self):
                self.name = "testInstruction"
                self.discriminator = [175, 175, 109, 31, 13, 152, 155, 237]

        ix = MockInstruction()
        # Should return the discriminator list
        assert get_instruction_discriminator(ix) == [175, 175, 109, 31, 13, 152, 155, 237]


class TestDefinedTypeHandling:
    """Test defined type handling for both formats."""

    def test_defined_type_old_format_string(self):
        """Test old format where defined is a string."""
        # Old format: defined is directly a string
        defined = "UpdateConfigMode"
        assert get_defined_type_name(defined) == "UpdateConfigMode"

    def test_defined_type_new_format_object(self):
        """Test new format where defined is an object with name field."""
        # New format: defined is an object with 'name' field
        class MockDefinedType:
            def __init__(self):
                self.name = "FooStruct"

        defined = MockDefinedType()
        assert get_defined_type_name(defined) == "FooStruct"

    def test_defined_type_new_format_dict(self):
        """Test new format where defined is a dict with name field."""
        # New format: defined could be a dict (from JSON parsing)
        defined = {"name": "BarEnum"}
        assert get_defined_type_name(defined) == "BarEnum"


class TestAccountFieldCompatibility:
    """Test account field name compatibility."""

    def test_account_writable_old_format(self):
        """Test old format using isMut field."""
        class MockAccount:
            def __init__(self):
                self.isMut = True

        acc = MockAccount()
        assert get_account_writable(acc) == True

        acc.isMut = False
        assert get_account_writable(acc) == False

    def test_account_writable_new_format(self):
        """Test new format using writable field."""
        class MockAccount:
            def __init__(self):
                self.writable = True

        acc = MockAccount()
        assert get_account_writable(acc) == True

        acc.writable = False
        assert get_account_writable(acc) == False

    def test_account_signer_old_format(self):
        """Test old format using isSigner field."""
        class MockAccount:
            def __init__(self):
                self.isSigner = True

        acc = MockAccount()
        assert get_account_signer(acc) == True

        acc.isSigner = False
        assert get_account_signer(acc) == False

    def test_account_signer_new_format(self):
        """Test new format using signer field."""
        class MockAccount:
            def __init__(self):
                self.signer = True

        acc = MockAccount()
        assert get_account_signer(acc) == True

        acc.signer = False
        assert get_account_signer(acc) == False

    def test_account_no_fields_defaults_to_false(self):
        """Test that missing fields default to False."""
        class MockAccount:
            def __init__(self):
                self.name = "test"

        acc = MockAccount()
        assert get_account_writable(acc) == False
        assert get_account_signer(acc) == False


class TestCoderCompatibility:
    """Test that coders work with both IDL formats."""

    @pytest.mark.asyncio
    async def test_accounts_coder_with_old_format(self):
        """Test AccountsCoder with old format IDL."""
        # Load an old format IDL
        idl_path = Path(__file__).parent / "idls" / "basic_1.json"
        if idl_path.exists():
            with open(idl_path) as f:
                idl_data = json.load(f)
                idl = Idl.from_json(json.dumps(idl_data))

                # Create AccountsCoder
                coder = AccountsCoder(idl)

                # Verify discriminators were calculated
                assert len(coder.acc_name_to_discriminator) > 0
                for name, disc in coder.acc_name_to_discriminator.items():
                    assert isinstance(disc, bytes)
                    assert len(disc) == 8

    @pytest.mark.asyncio
    async def test_instruction_coder_with_old_format(self):
        """Test InstructionCoder with old format IDL."""
        # Load an old format IDL
        idl_path = Path(__file__).parent / "idls" / "basic_1.json"
        if idl_path.exists():
            with open(idl_path) as f:
                idl_data = json.load(f)
                idl = Idl.from_json(json.dumps(idl_data))

                # Create InstructionCoder
                coder = InstructionCoder(idl)

                # Verify sighashes were calculated
                assert len(coder.sighashes) > 0
                for name, sighash in coder.sighashes.items():
                    assert isinstance(sighash, bytes)
                    assert len(sighash) == 8


class TestEndToEndCompatibility:
    """Test end-to-end compatibility with real IDL files."""

    @pytest.mark.asyncio
    async def test_program_with_old_format_idl(self):
        """Test creating a Program with an old format IDL."""
        idl_path = Path(__file__).parent / "idls" / "basic_1.json"
        if idl_path.exists():
            with open(idl_path) as f:
                idl_data = json.load(f)

            # Create a mock provider
            provider = Provider.readonly()

            # Create program with old format IDL
            program_id = Pubkey.new_unique()
            program = Program(
                idl_data,
                program_id,
                provider
            )

            # Verify program was created successfully
            assert program is not None
            assert program.program_id == program_id

            # Verify coders were initialized
            assert program.coder is not None
            assert program.coder.accounts is not None
            assert program.coder.instruction is not None


from __future__ import annotations
import typing
from dataclasses import dataclass
from anchorpy.borsh_extension import EnumForCodegen
import borsh_construct as borsh


class MintTokensJSON(typing.TypedDict):
    kind: typing.Literal["MintTokens"]


class FreezeAccountJSON(typing.TypedDict):
    kind: typing.Literal["FreezeAccount"]


class AccountOwnerJSON(typing.TypedDict):
    kind: typing.Literal["AccountOwner"]


class CloseAccountJSON(typing.TypedDict):
    kind: typing.Literal["CloseAccount"]


@dataclass
class MintTokens:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "MintTokens"

    @classmethod
    def to_json(cls) -> MintTokensJSON:
        return MintTokensJSON(
            kind="MintTokens",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "MintTokens": {},
        }


@dataclass
class FreezeAccount:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "FreezeAccount"

    @classmethod
    def to_json(cls) -> FreezeAccountJSON:
        return FreezeAccountJSON(
            kind="FreezeAccount",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "FreezeAccount": {},
        }


@dataclass
class AccountOwner:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "AccountOwner"

    @classmethod
    def to_json(cls) -> AccountOwnerJSON:
        return AccountOwnerJSON(
            kind="AccountOwner",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "AccountOwner": {},
        }


@dataclass
class CloseAccount:
    discriminator: typing.ClassVar = 3
    kind: typing.ClassVar = "CloseAccount"

    @classmethod
    def to_json(cls) -> CloseAccountJSON:
        return CloseAccountJSON(
            kind="CloseAccount",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "CloseAccount": {},
        }


AuthorityTypeKind = typing.Union[MintTokens, FreezeAccount, AccountOwner, CloseAccount]
AuthorityTypeJSON = typing.Union[
    MintTokensJSON, FreezeAccountJSON, AccountOwnerJSON, CloseAccountJSON
]


def from_decoded(obj: dict) -> AuthorityTypeKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "MintTokens" in obj:
        return MintTokens()
    if "FreezeAccount" in obj:
        return FreezeAccount()
    if "AccountOwner" in obj:
        return AccountOwner()
    if "CloseAccount" in obj:
        return CloseAccount()
    raise ValueError("Invalid enum object")


def from_json(obj: AuthorityTypeJSON) -> AuthorityTypeKind:
    if obj["kind"] == "MintTokens":
        return MintTokens()
    if obj["kind"] == "FreezeAccount":
        return FreezeAccount()
    if obj["kind"] == "AccountOwner":
        return AccountOwner()
    if obj["kind"] == "CloseAccount":
        return CloseAccount()
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "MintTokens" / borsh.CStruct(),
    "FreezeAccount" / borsh.CStruct(),
    "AccountOwner" / borsh.CStruct(),
    "CloseAccount" / borsh.CStruct(),
)



"""This module contains code handling the Context object."""
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

from anchorpy3_core.idl import IdlInstruction
from pyheck import snake
from solana.rpc.types import TxOpts
from solders.instruction import AccountMeta, Instruction
from solders.keypair import Keypair

from anchorpy.error import ArgsError

# should be Dict[str, Union[Pubkey, Accounts]]
# but mypy doesn't support recursive types
Accounts = Dict[str, Any]


@dataclass
class Context:
    """Context provides all non-argument inputs for generating Anchor transactions.

    Attributes:
        accounts: The accounts used in the instruction context.
        remaining_accounts: All accounts to pass into an instruction *after* the main
        `accounts`. This can be used for optional or otherwise unknown accounts.
        signers: Accounts that must sign a given transaction.
        pre_instructions: Instructions to run *before* a given method. Often this is
            used, for example to create accounts prior to executing a method.
        post_instructions: Instructions to run *after* a given method. Often this is
            used, for example to close accounts prior to executing a method.
        options: Commitment parameters to use for a transaction.

    """

    # For some reason mkdocstrings doesn't understand the full type hint
    # here if we use list[Instruction] instead of typing.List.
    # Weirdly there are other places where it understands list[whatever].

    accounts: Accounts = field(default_factory=dict)
    remaining_accounts: List[AccountMeta] = field(default_factory=list)
    signers: List[Keypair] = field(default_factory=list)
    pre_instructions: List[Instruction] = field(default_factory=list)
    post_instructions: List[Instruction] = field(default_factory=list)
    options: Optional[TxOpts] = None


def _check_args_length(
    idl_ix: IdlInstruction,
    args: Tuple,
) -> None:
    """Check that the correct number of args is passed to the RPC function.

    Args:
        idl_ix: The IDL instruction object.
        args: The instruction arguments.

    Raises:
        ArgsError: If the correct number of args is not parsed.
    """
    if len(args) != len(idl_ix.args):
        expected_arg_names = [snake(arg.name) for arg in idl_ix.args]
        raise ArgsError(
            f"Provided incorrect number of args to instruction={snake(idl_ix.name)}. "
            f"Expected {expected_arg_names}",
            f"Received {args}",
        )


EMPTY_CONTEXT = Context()



#!/usr/bin/env python3
"""Test loopscale_v2.json with the updated AnchorPy implementation."""

import json
import tempfile
from pathlib import Path
from anchorpy3_core.idl import Idl
from anchorpy import Program, Provider
from anchorpy.clientgen.accounts import gen_accounts
from anchorpy.clientgen.errors import gen_errors
from anchorpy.clientgen.instructions import gen_instructions
from anchorpy.clientgen.program_id import gen_program_id
from anchorpy.clientgen.types import gen_types

def test_loopscale_v2():
    """Test loopscale_v2.json IDL parsing and client generation."""

    print("="*60)
    print("Testing loopscale_v2.json")
    print("="*60)

    # Load the IDL
    idl_path = Path(__file__).parent.parent / "loopscale_v2.json"
    print(f"Loading IDL from: {idl_path}")

    with open(idl_path) as f:
        idl_json = json.load(f)

    # Test 1: Parse as Idl object
    print("\n1. Testing IDL parsing...")
    try:
        idl_obj = Idl.from_json(json.dumps(idl_json))
        print(f"   ✅ Successfully parsed IDL")
        print(f"   - Version: {idl_json.get('version', 'N/A')}")
        print(f"   - Address: {idl_json.get('address', 'N/A')}")
        if idl_obj.metadata:
            print(f"   - Metadata spec: {getattr(idl_obj.metadata, 'spec', 'N/A')}")
        print(f"   - Accounts: {len(idl_obj.accounts) if idl_obj.accounts else 0}")
        print(f"   - Instructions: {len(idl_obj.instructions) if idl_obj.instructions else 0}")
        print(f"   - Types: {len(idl_obj.types) if idl_obj.types else 0}")
        print(f"   - Events: {len(idl_obj.events) if idl_obj.events else 0}")
        print(f"   - Errors: {len(idl_obj.errors) if idl_obj.errors else 0}")

        # Check for discriminators (new format indicator)
        if idl_obj.accounts and len(idl_obj.accounts) > 0:
            first_acc = idl_obj.accounts[0]
            if hasattr(first_acc, 'discriminator') or (isinstance(first_acc, dict) and 'discriminator' in first_acc):
                print("   ✅ New format detected (has discriminators)")
            else:
                print("   ℹ️ Old format detected (no discriminators)")

    except Exception as e:
        print(f"   ❌ Failed to parse IDL: {e}")
        return False

    # Test 2: Create Program object
    print("\n2. Testing Program creation...")
    try:
        provider = Provider.local()
        program_id = idl_json.get('address', 'So11111111111111111111111111111111111111112')
        program = Program(idl_json, program_id, provider)
        print(f"   ✅ Successfully created Program object")
        print(f"   - Program ID: {program.program_id}")
    except Exception as e:
        print(f"   ❌ Failed to create Program: {e}")
        return False

    # Test 3: Generate client code
    print("\n3. Testing client code generation...")
    program_id = idl_json.get('address')
    if not program_id:
        metadata = idl_json.get('metadata', {})
        program_id = metadata.get('address')
    if not program_id:
        program_id = 'So11111111111111111111111111111111111111112'

    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = Path(tmpdir) / "generated"
        output_path.mkdir()

        try:
            # Generate each component
            gen_program_id(program_id, output_path)
            print("   ✅ Generated program_id.py")

            gen_errors(idl_obj, output_path)
            print("   ✅ Generated errors.py")

            gen_instructions(idl_obj, output_path, gen_pdas=False)
            print("   ✅ Generated instructions")

            gen_types(idl_obj, output_path)
            print("   ✅ Generated types")

            gen_accounts(idl_obj, output_path)
            print("   ✅ Generated accounts")

            # Count generated files
            account_files = list((output_path / "accounts").glob("*.py")) if (output_path / "accounts").exists() else []
            instruction_files = list((output_path / "instructions").glob("*.py")) if (output_path / "instructions").exists() else []
            type_files = list((output_path / "types").glob("*.py")) if (output_path / "types").exists() else []

            print(f"\n   Summary:")
            print(f"   - {len(account_files)} account files")
            print(f"   - {len(instruction_files)} instruction files")
            print(f"   - {len(type_files)} type files")

            # List some account names for verification
            if account_files:
                print(f"\n   Sample accounts generated:")
                for f in sorted(account_files)[:5]:
                    print(f"     - {f.stem}")
                if len(account_files) > 5:
                    print(f"     ... and {len(account_files) - 5} more")

            return True

        except Exception as e:
            print(f"   ❌ Failed to generate client code: {e}")
            import traceback
            traceback.print_exc()
            return False

if __name__ == "__main__":
    success = test_loopscale_v2()

    print("\n" + "="*60)
    if success:
        print("✅ SUCCESS: loopscale_v2.json fully supported!")
    else:
        print("❌ FAILED: Issues with loopscale_v2.json")
    print("="*60)


from __future__ import annotations
import typing
from dataclasses import dataclass
from solders.pubkey import Pubkey
from anchorpy.borsh_extension import EnumForCodegen, BorshPubkey
import borsh_construct as borsh


class WonJSONValue(typing.TypedDict):
    winner: str


class WonValue(typing.TypedDict):
    winner: Pubkey


class ActiveJSON(typing.TypedDict):
    kind: typing.Literal["Active"]


class TieJSON(typing.TypedDict):
    kind: typing.Literal["Tie"]


class WonJSON(typing.TypedDict):
    value: WonJSONValue
    kind: typing.Literal["Won"]


@dataclass
class Active:
    discriminator: typing.ClassVar = 0
    kind: typing.ClassVar = "Active"

    @classmethod
    def to_json(cls) -> ActiveJSON:
        return ActiveJSON(
            kind="Active",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Active": {},
        }


@dataclass
class Tie:
    discriminator: typing.ClassVar = 1
    kind: typing.ClassVar = "Tie"

    @classmethod
    def to_json(cls) -> TieJSON:
        return TieJSON(
            kind="Tie",
        )

    @classmethod
    def to_encodable(cls) -> dict:
        return {
            "Tie": {},
        }


@dataclass
class Won:
    discriminator: typing.ClassVar = 2
    kind: typing.ClassVar = "Won"
    value: WonValue

    def to_json(self) -> WonJSON:
        return WonJSON(
            kind="Won",
            value={
                "winner": str(self.value["winner"]),
            },
        )

    def to_encodable(self) -> dict:
        return {
            "Won": {
                "winner": self.value["winner"],
            },
        }


GameStateKind = typing.Union[Active, Tie, Won]
GameStateJSON = typing.Union[ActiveJSON, TieJSON, WonJSON]


def from_decoded(obj: dict) -> GameStateKind:
    if not isinstance(obj, dict):
        raise ValueError("Invalid enum object")
    if "Active" in obj:
        return Active()
    if "Tie" in obj:
        return Tie()
    if "Won" in obj:
        val = obj["Won"]
        return Won(
            WonValue(
                winner=val["winner"],
            )
        )
    raise ValueError("Invalid enum object")


def from_json(obj: GameStateJSON) -> GameStateKind:
    if obj["kind"] == "Active":
        return Active()
    if obj["kind"] == "Tie":
        return Tie()
    if obj["kind"] == "Won":
        won_json_value = typing.cast(WonJSONValue, obj["value"])
        return Won(
            WonValue(
                winner=Pubkey.from_string(won_json_value["winner"]),
            )
        )
    kind = obj["kind"]
    raise ValueError(f"Unrecognized enum kind: {kind}")


layout = EnumForCodegen(
    "Active" / borsh.CStruct(),
    "Tie" / borsh.CStruct(),
    "Won" / borsh.CStruct("winner" / BorshPubkey),
)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
from ..program_id import PROGRAM_ID


class InitializeImmutableOwnerAccounts(typing.TypedDict):
    account: Pubkey


def initialize_immutable_owner(
    accounts: InitializeImmutableOwnerAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True)
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b'\x8d2\x0f,\xc3\xf7"<'
    encoded_args = b""
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.sysvar import RENT
from solders.instruction import Instruction, AccountMeta
from anchorpy.borsh_extension import BorshPubkey
import borsh_construct as borsh
from ..program_id import PROGRAM_ID


class InitializeAccount2Args(typing.TypedDict):
    owner: Pubkey


layout = borsh.CStruct("owner" / BorshPubkey)


class InitializeAccount2Accounts(typing.TypedDict):
    account: Pubkey
    mint: Pubkey


def initialize_account2(
    args: InitializeAccount2Args,
    accounts: InitializeAccount2Accounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["account"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["mint"], is_signer=False, is_writable=False),
        AccountMeta(pubkey=RENT, is_signer=False, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\x08\xb6\x95\x90\xb9\x1f\xd1i"
    encoded_args = layout.build(
        {
            "owner": args["owner"],
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



#!/usr/bin/env python3
"""Test client generation for both old and new IDL formats."""

import json
import tempfile
from pathlib import Path
from anchorpy import Program, Provider
from anchorpy3_core.idl import Idl
from anchorpy.clientgen import generate
import sys

def test_clientgen(idl_name: str):
    """Test client generation for a specific IDL."""
    print(f"\n{'='*60}")
    print(f"Testing client generation for {idl_name}")
    print(f"{'='*60}")

    # Load the IDL
    idl_path = Path(__file__).parent.parent / idl_name
    print(f"Loading IDL from: {idl_path}")

    with open(idl_path) as f:
        idl_json = json.load(f)

    # Parse as Idl object
    idl = Idl.from_json(json.dumps(idl_json))

    # Create a temporary directory for generated code
    with tempfile.TemporaryDirectory() as tmpdir:
        output_path = Path(tmpdir) / "generated"
        output_path.mkdir()

        print(f"Generating client code to: {output_path}")

        try:
            # Generate the client code
            generate(
                idl=idl,
                root=output_path,
                program_id=idl_json.get('address', 'So11111111111111111111111111111111111111112')
            )
            print(f"✅ Successfully generated client code for {idl_name}")

            # List generated files
            print("\nGenerated files:")
            for file in sorted(output_path.rglob("*.py")):
                rel_path = file.relative_to(output_path)
                print(f"  - {rel_path}")

            # Try to import the generated module (basic validation)
            if (output_path / "__init__.py").exists():
                print("\n✅ Generated __init__.py exists")

            # Check if accounts were generated
            accounts_dir = output_path / "accounts"
            if accounts_dir.exists():
                account_files = list(accounts_dir.glob("*.py"))
                print(f"\n✅ Generated {len(account_files)} account files")

            # Check if instructions were generated
            instructions_dir = output_path / "instructions"
            if instructions_dir.exists():
                instruction_files = list(instructions_dir.glob("*.py"))
                print(f"✅ Generated {len(instruction_files)} instruction files")

            # Check if types were generated
            types_dir = output_path / "types"
            if types_dir.exists():
                type_files = list(types_dir.glob("*.py"))
                print(f"✅ Generated {len(type_files)} type files")

            return True

        except Exception as e:
            print(f"❌ Failed to generate client code: {e}")
            import traceback
            traceback.print_exc()
            return False

# Test all three IDLs
idls = [
    ("kamino_lend_v4.json", "Old format IDL"),
    ("adrena.json", "Old format IDL"),
    ("loopscale_v1.json", "New format IDL (v0.1.0)")
]

results = []
for idl_file, description in idls:
    print(f"\nTesting {idl_file} ({description})...")
    success = test_clientgen(idl_file)
    results.append((idl_file, success))

# Print summary
print(f"\n{'='*60}")
print("SUMMARY")
print(f"{'='*60}")
for idl_file, success in results:
    status = "✅ PASSED" if success else "❌ FAILED"
    print(f"{idl_file:30} {status}")

# Exit with appropriate code
if all(success for _, success in results):
    print("\n🎉 All client generation tests passed!")
    sys.exit(0)
else:
    print("\n⚠️ Some client generation tests failed.")
    sys.exit(1)


import typing
from . import bar_struct
from .bar_struct import BarStruct, BarStructJSON
from . import foo_struct
from .foo_struct import FooStruct, FooStructJSON
from . import foo_enum
from .foo_enum import FooEnumKind, FooEnumJSON
from . import u8_array
from .u8_array import U8Array



import json
from filecmp import dircmp
from pathlib import Path

from anchorpy.cli import client_gen
from py.path import local
from solana.rpc.core import RPCException
from solders.pubkey import Pubkey
from solders.rpc.errors import SendTransactionPreflightFailureMessage
from solders.rpc.responses import SimulateTransactionResp

from tests.client_gen.example_program_gen.accounts import State
from tests.client_gen.example_program_gen.errors import from_tx_error
from tests.client_gen.example_program_gen.errors.anchor import InvalidProgramId
from tests.client_gen.example_program_gen.types import BarStruct, FooStruct
from tests.client_gen.example_program_gen.types.foo_enum import (
    Named,
    NamedValue,
    NoFields,
    Struct,
    Unnamed,
)


def test_quarry_mine(tmpdir: local) -> None:
    proj_dir = Path(tmpdir)
    out_dir = proj_dir / "generated"
    idl_path = Path("tests/idls/quarry_mine.json")
    client_gen(idl_path, out_dir, "placeholder")


def test_merkle_distributor(tmpdir: local) -> None:
    proj_dir = Path(tmpdir)
    out_dir = proj_dir / "generated"
    idl_path = Path("tests/idls/merkle_distributor.json")
    client_gen(idl_path, out_dir, "placeholder")


def test_null_err_when_cpi_fails() -> None:
    to_dump = {
        "jsonrpc": "2.0",
        "error": {
            "code": -32002,
            "message": "",
            "data": {
                "err": {"InstructionError": [0, {"Custom": 3}]},
                "logs": [
                    "Program 3rTQ3R4B2PxZrAyx7EUefySPgZY8RhJf16cZajbmrzp8 invoke [1]",
                    "Program log: Instruction: CauseError",
                    "Program 11111111111111111111111111111111 invoke [2]",
                    "Allocate: requested 1000000000000000000, max allowed 10485760",
                    "Program 11111111111111111111111111111111 failed: custom program error: 0x3",
                    "Program 3rTQ3R4B2PxZrAyx7EUefySPgZY8RhJf16cZajbmrzp8 consumed 7958 of 1400000 compute units",
                    "Program 3rTQ3R4B2PxZrAyx7EUefySPgZY8RhJf16cZajbmrzp8 failed: custom program error: 0x3",
                ],
            },
        },
    }
    raw = json.dumps(to_dump)
    parsed = SimulateTransactionResp.from_json(raw)
    assert isinstance(parsed, SendTransactionPreflightFailureMessage)
    err_mock = RPCException(parsed)
    assert from_tx_error(err_mock) is None


def test_parses_anchor_error() -> None:
    to_dump = {
        "jsonrpc": "2.0",
        "error": {
            "code": -32002,
            "message": "",
            "data": {
                "err": {"InstructionError": [0, {"Custom": 3008}]},
                "logs": [
                    "Program 3rTQ3R4B2PxZrAyx7EUefySPgZY8RhJf16cZajbmrzp8 invoke [1]",
                    "Program log: Instruction: CauseError",
                    "Program log: AnchorError caused by account: system_program. Error Code: InvalidProgramId. Error Number: 3008. Error Message: Program ID was not as expected.",
                    "Program log: Left:",
                    "Program log: 24S58Cp5Myf6iGx4umBNd7RgDrZ9nkKzvkfFHBMDomNa",
                    "Program log: Right:",
                    "Program log: 11111111111111111111111111111111",
                    "Program 3rTQ3R4B2PxZrAyx7EUefySPgZY8RhJf16cZajbmrzp8 consumed 5043 of 1400000 compute units",
                    "Program 3rTQ3R4B2PxZrAyx7EUefySPgZY8RhJf16cZajbmrzp8 failed: custom program error: 0xbc0",
                ],
            },
        },
    }
    raw = json.dumps(to_dump)
    parsed = SimulateTransactionResp.from_json(raw)
    assert isinstance(parsed, SendTransactionPreflightFailureMessage)
    err_mock = RPCException(parsed)
    assert isinstance(from_tx_error(err_mock), InvalidProgramId)


def test_json() -> None:
    vec_struct_field = [
        FooStruct(
            field1=5,
            field2=6,
            nested=BarStruct(
                some_field=True,
                other_field=15,
            ),
            vec_nested=[
                BarStruct(
                    some_field=True,
                    other_field=13,
                ),
            ],
            option_nested=None,
            enum_field=Unnamed(
                (
                    False,
                    111,
                    BarStruct(
                        some_field=False,
                        other_field=11,
                    ),
                )
            ),
        ),
    ]
    option_struct_field = FooStruct(
        field1=8,
        field2=9,
        nested=BarStruct(
            some_field=True,
            other_field=17,
        ),
        vec_nested=[
            BarStruct(
                some_field=True,
                other_field=10,
            ),
        ],
        option_nested=BarStruct(
            some_field=False,
            other_field=99,
        ),
        enum_field=NoFields(),
    )
    struct_field = FooStruct(
        field1=11,
        field2=12,
        nested=BarStruct(
            some_field=False,
            other_field=177,
        ),
        vec_nested=[
            BarStruct(
                some_field=True,
                other_field=15,
            ),
        ],
        option_nested=BarStruct(
            some_field=True,
            other_field=75,
        ),
        enum_field=NoFields(),
    )
    enum_field1 = Unnamed(
        (
            False,
            157,
            BarStruct(
                some_field=True,
                other_field=193,
            ),
        )
    )
    enum_field2 = Named(
        NamedValue(
            bool_field=False,
            u8_field=77,
            nested=BarStruct(
                some_field=True,
                other_field=100,
            ),
        )
    )
    enum_field3 = Struct(
        (
            BarStruct(
                some_field=False,
                other_field=122,
            ),
        )
    )
    state = State(
        bool_field=True,
        u8_field=255,
        i8_field=-120,
        u16_field=62000,
        i16_field=-31000,
        u32_field=123456789,
        i32_field=-123456789,
        f32_field=123456.5,
        u64_field=9223372036854775805,
        i64_field=4611686018427387910,
        f64_field=1234567891.35,
        u128_field=170141183460469231731687303715884105760,
        i128_field=-85070591730234615865843651857942052897,
        bytes_field=bytes([1, 255]),
        string_field="a string",
        pubkey_field=Pubkey.from_string("EPZP2wrcRtMxrAPJCXVEQaYD9eH7fH7h12YqKDcd4aS7"),
        vec_field=[10, 1234567890123456],
        vec_struct_field=vec_struct_field,
        option_field=None,
        option_struct_field=option_struct_field,
        struct_field=struct_field,
        array_field=[True, False],
        enum_field1=enum_field1,
        enum_field2=enum_field2,
        enum_field3=enum_field3,
        enum_field4=NoFields(),
    )
    state_json = state.to_json()
    expected = {
        "bool_field": True,
        "u8_field": 255,
        "i8_field": -120,
        "u16_field": 62000,
        "i16_field": -31000,
        "u32_field": 123456789,
        "i32_field": -123456789,
        "f32_field": 123456.5,
        "u64_field": 9223372036854775805,
        "i64_field": 4611686018427387910,
        "f64_field": 1234567891.35,
        "u128_field": 170141183460469231731687303715884105760,
        "i128_field": -85070591730234615865843651857942052897,
        "bytes_field": [1, 255],
        "string_field": "a string",
        "pubkey_field": "EPZP2wrcRtMxrAPJCXVEQaYD9eH7fH7h12YqKDcd4aS7",
        "vec_field": [10, 1234567890123456],
        "vec_struct_field": [
            {
                "field1": 5,
                "field2": 6,
                "nested": {
                    "some_field": True,
                    "other_field": 15,
                },
                "vec_nested": [
                    {
                        "some_field": True,
                        "other_field": 13,
                    },
                ],
                "option_nested": None,
                "enum_field": {
                    "kind": "Unnamed",
                    "value": (
                        False,
                        111,
                        {
                            "some_field": False,
                            "other_field": 11,
                        },
                    ),
                },
            },
        ],
        "option_field": None,
        "option_struct_field": {
            "field1": 8,
            "field2": 9,
            "nested": {
                "some_field": True,
                "other_field": 17,
            },
            "vec_nested": [
                {
                    "some_field": True,
                    "other_field": 10,
                },
            ],
            "option_nested": {
                "some_field": False,
                "other_field": 99,
            },
            "enum_field": {
                "kind": "NoFields",
            },
        },
        "struct_field": {
            "field1": 11,
            "field2": 12,
            "nested": {
                "some_field": False,
                "other_field": 177,
            },
            "vec_nested": [
                {
                    "some_field": True,
                    "other_field": 15,
                },
            ],
            "option_nested": {
                "some_field": True,
                "other_field": 75,
            },
            "enum_field": {
                "kind": "NoFields",
            },
        },
        "array_field": [True, False],
        "enum_field1": {
            "kind": "Unnamed",
            "value": (
                False,
                157,
                {
                    "some_field": True,
                    "other_field": 193,
                },
            ),
        },
        "enum_field2": {
            "kind": "Named",
            "value": {
                "bool_field": False,
                "u8_field": 77,
                "nested": {
                    "some_field": True,
                    "other_field": 100,
                },
            },
        },
        "enum_field3": {
            "kind": "Struct",
            "value": (
                {
                    "some_field": False,
                    "other_field": 122,
                },
            ),
        },
        "enum_field4": {
            "kind": "NoFields",
        },
    }
    assert state_json == expected
    state_from_json = State.from_json(state_json)
    assert state_from_json == state


def has_differences(dcmp: dircmp) -> bool:
    differences = dcmp.left_only + dcmp.right_only + dcmp.diff_files
    if differences:
        return True
    return any([has_differences(subdcmp) for subdcmp in dcmp.subdirs.values()])


def test_generated_as_expected(project_dir: Path) -> None:
    dcmp = dircmp(project_dir, "tests/client_gen/example_program_gen")
    assert not has_differences(dcmp)



"""The Python Anchor client."""
from contextlib import suppress as __suppress

from anchorpy3_core.idl import Idl

from anchorpy import error, utils
from anchorpy.coder.coder import AccountsCoder, Coder, EventCoder, InstructionCoder
from anchorpy.idl import IdlProgramAccount
from anchorpy.program.common import (
    Event,
    NamedInstruction,
    translate_address,
    validate_accounts,
)
from anchorpy.program.context import Context
from anchorpy.program.core import Program
from anchorpy.program.event import EventParser
from anchorpy.program.namespace.account import AccountClient, ProgramAccount
from anchorpy.program.namespace.simulate import SimulateResponse
from anchorpy.provider import Provider, Wallet
from anchorpy.workspace import WorkspaceType, close_workspace, create_workspace

__has_pytest = False
with __suppress(ImportError):
    from anchorpy.pytest_plugin import (
        localnet_fixture,
        workspace_fixture,
    )

    __has_pytest = True

__all_core = [
    "Program",
    "Provider",
    "Context",
    "create_workspace",
    "close_workspace",
    "Idl",
    "WorkspaceType",
    "Wallet",
    "Coder",
    "InstructionCoder",
    "EventCoder",
    "AccountsCoder",
    "NamedInstruction",
    "IdlProgramAccount",
    "Event",
    "translate_address",
    "validate_accounts",
    "AccountClient",
    "ProgramAccount",
    "EventParser",
    "SimulateResponse",
    "error",
    "utils",
]

__all__ = (
    [
        *__all_core,
        "localnet_fixture",
        "workspace_fixture",
    ]
    if __has_pytest
    else __all_core
)

__version__ = "0.3.0"



"""Provides the `AccountClient` class."""
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union

from anchorpy3_core.idl import Idl, IdlTypeDefinition
from based58 import b58encode
from construct import Container
from solana.rpc.commitment import Commitment
from solana.rpc.types import MemcmpOpts
from solders.instruction import Instruction
from solders.keypair import Keypair
from solders.pubkey import Pubkey
from solders.system_program import CreateAccountParams, create_account

from anchorpy.coder.accounts import ACCOUNT_DISCRIMINATOR_SIZE
from anchorpy.coder.coder import Coder
from anchorpy.coder.common import _account_size
from anchorpy.error import AccountDoesNotExistError, AccountInvalidDiscriminator
from anchorpy.provider import Provider
from anchorpy.utils.rpc import get_multiple_accounts


def _build_account(
    idl: Idl,
    coder: Coder,
    program_id: Pubkey,
    provider: Provider,
) -> Dict[str, "AccountClient"]:
    """Generate the `.account` namespace.

    Args:
        idl: The parsed Idl object.
        coder: The program's coder object.
        program_id: The program ID.
        provider: The Provider instance.

    Returns:
        Mapping of account name to `AccountClient` instance.
    """
    accounts_fns = {}
    for idl_account in idl.accounts:
        account_client = AccountClient(idl, idl_account, coder, program_id, provider)
        # Get account name (handle both object and dict)
        acc_name = idl_account.name if hasattr(idl_account, 'name') else idl_account.get('name')
        accounts_fns[acc_name] = account_client
    return accounts_fns


@dataclass
class ProgramAccount:
    """Deserialized account owned by a program."""

    public_key: Pubkey
    account: Container


class AccountClient(object):
    """Provides methods for fetching and creating accounts."""

    def __init__(
        self,
        idl: Idl,
        idl_account: IdlTypeDefinition,
        coder: Coder,
        program_id: Pubkey,
        provider: Provider,
    ):
        """Init.

        Args:
            idl: the parsed IDL object.
            idl_account: the account definition from the IDL.
            coder: The program's Coder object.
            program_id: the program ID.
            provider: The Provider object for the Program.
        """
        self._idl_account = idl_account
        self._program_id = program_id
        self._provider = provider
        self._coder = coder
        self._size = ACCOUNT_DISCRIMINATOR_SIZE + _account_size(idl, idl_account)
        # Get account name (handle both object and dict)
        acc_name = idl_account.name if hasattr(idl_account, 'name') else idl_account.get('name')
        # Get discriminator from coder (which handles both old and new formats)
        self._discriminator = self._coder.accounts.acc_name_to_discriminator[acc_name]

    async def fetch(
        self, address: Pubkey, commitment: Optional[Commitment] = None
    ) -> Container[Any]:
        """Return a deserialized account.

        Args:
            address: The address of the account to fetch.
            commitment: Bank state to query.


        Raises:
            AccountDoesNotExistError: If the account doesn't exist.
            AccountInvalidDiscriminator: If the discriminator doesn't match the IDL.
        """
        account_info = await self._provider.connection.get_account_info(
            address,
            encoding="base64",
            commitment=commitment,
        )
        if not account_info.value:
            raise AccountDoesNotExistError(f"Account {address} does not exist")
        data = account_info.value.data
        # Use cached discriminator (which supports both old and new formats)
        if self._discriminator != data[:ACCOUNT_DISCRIMINATOR_SIZE]:
            msg = f"Account {address} has an invalid discriminator"
            raise AccountInvalidDiscriminator(msg)
        return self._coder.accounts.decode(data)

    async def fetch_multiple(
        self,
        addresses: List[Pubkey],
        batch_size: int = 300,
        commitment: Optional[Commitment] = None,
    ) -> list[Optional[Container[Any]]]:
        """Return multiple deserialized accounts.

        Accounts not found or with wrong discriminator are returned as None.

        Args:
            addresses: The addresses of the accounts to fetch.
            batch_size: The number of `getMultipleAccounts` objects to send
                in each HTTP request.
            commitment: Bank state to query.
        """
        accounts = await get_multiple_accounts(
            self._provider.connection,
            addresses,
            batch_size=batch_size,
            commitment=commitment,
        )
        # Use cached discriminator (which supports both old and new formats)
        result: list[Optional[Container[Any]]] = []
        for account in accounts:
            if account is None:
                result.append(None)
            elif self._discriminator == account.account.data[:8]:
                result.append(self._coder.accounts.decode(account.account.data))
            else:
                result.append(None)
        return result

    async def create_instruction(
        self,
        signer: Keypair,
        size_override: int = 0,
    ) -> Instruction:
        """Return an instruction for creating this account.

        Args:
            signer: [description]
            size_override: Optional override for the account size. Defaults to 0.

        Returns:
            The instruction to create the account.
        """
        space = size_override if size_override else self._size
        mbre_resp = (
            await self._provider.connection.get_minimum_balance_for_rent_exemption(
                space
            )
        )
        return create_account(
            CreateAccountParams(
                from_pubkey=self._provider.wallet.public_key,
                to_pubkey=signer.pubkey(),
                space=space,
                lamports=mbre_resp.value,
                owner=self._program_id,
            )
        )

    async def all(  # noqa: A003
        self,
        buffer: Optional[bytes] = None,
        filters: Optional[List[Union[int, MemcmpOpts]]] = None,
    ) -> list[ProgramAccount]:
        """Return all instances of this account type for the program.

        Args:
            buffer: bytes filter to append to the discriminator.
            filters: (optional) Options to compare a provided series of bytes with
                program account data at a particular offset.
                Note: an int entry is converted to a `dataSize` filter.
        """
        all_accounts = []
        # Use cached discriminator (which supports both old and new formats)
        to_encode = self._discriminator if buffer is None else self._discriminator + buffer
        bytes_arg = b58encode(to_encode).decode("ascii")
        base_memcmp_opt = MemcmpOpts(
            offset=0,
            bytes=bytes_arg,
        )
        filters_to_use = [base_memcmp_opt] + ([] if filters is None else filters)
        resp = await self._provider.connection.get_program_accounts(
            self._program_id,
            encoding="base64",
            commitment=self.provider.connection._commitment,
            filters=filters_to_use,
        )
        for r in resp.value:
            account_data = r.account.data
            all_accounts.append(
                ProgramAccount(
                    public_key=r.pubkey,
                    account=self._coder.accounts.decode(account_data),
                ),
            )
        return all_accounts

    @property
    def size(self) -> int:
        """Return the number of bytes in this account."""
        return self._size

    @property
    def program_id(self) -> Pubkey:
        """Return the program ID owning all accounts."""
        return self._program_id

    @property
    def provider(self) -> Provider:
        """Return the client's wallet and network provider."""
        return self._provider

    @property
    def coder(self) -> Coder:
        """Return the coder."""
        return self._coder



from __future__ import annotations
import typing
from solders.pubkey import Pubkey
from solders.instruction import Instruction, AccountMeta
import borsh_construct as borsh
from .. import types
from ..program_id import PROGRAM_ID


class PlayArgs(typing.TypedDict):
    tile: types.tile.Tile


layout = borsh.CStruct("tile" / types.tile.Tile.layout)


class PlayAccounts(typing.TypedDict):
    game: Pubkey
    player: Pubkey


def play(
    args: PlayArgs,
    accounts: PlayAccounts,
    program_id: Pubkey = PROGRAM_ID,
    remaining_accounts: typing.Optional[typing.List[AccountMeta]] = None,
) -> Instruction:
    keys: list[AccountMeta] = [
        AccountMeta(pubkey=accounts["game"], is_signer=False, is_writable=True),
        AccountMeta(pubkey=accounts["player"], is_signer=True, is_writable=False),
    ]
    if remaining_accounts is not None:
        keys += remaining_accounts
    identifier = b"\xd5\x9d\xc1\x8e\xe48\xf8\x96"
    encoded_args = layout.build(
        {
            "tile": args["tile"].to_encodable(),
        }
    )
    data = identifier + encoded_args
    return Instruction(program_id, data, keys)



import typing
import re
from solders.transaction_status import (
    InstructionErrorCustom,
    TransactionErrorInstructionError,
)
from solana.rpc.core import RPCException
from solders.rpc.errors import SendTransactionPreflightFailureMessage
from anchorpy.error import extract_code_and_logs
from ..program_id import PROGRAM_ID
from . import anchor


def from_code(code: int) -> typing.Optional[anchor.AnchorError]:
    return anchor.from_code(code)


error_re = re.compile(r"Program (\w+) failed: custom program error: (\w+)")


def from_tx_error(error: RPCException) -> typing.Optional[anchor.AnchorError]:
    err_info = error.args[0]
    extracted = extract_code_and_logs(err_info, PROGRAM_ID)
    if extracted is None:
        return None
    return from_code(extracted[0])



import os
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, cast

import typer
from anchorpy3_core.idl import Idl
from IPython import embed

from anchorpy import create_workspace
from anchorpy.clientgen.accounts import gen_accounts
from anchorpy.clientgen.errors import gen_errors
from anchorpy.clientgen.instructions import gen_instructions
from anchorpy.clientgen.program_id import gen_program_id
from anchorpy.clientgen.types import gen_types
from anchorpy.template import INIT_TESTS

app = typer.Typer()


@contextmanager
def _set_directory(path: Path):
    """Set the cwd within the context.

    Args:
        path (Path): The path to the cwd

    Yields:
        None
    """  # noqa: D202

    origin = Path().absolute()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(origin)


def _search_upwards_for_project_root() -> Path:
    """Search in the current dir and all directories above it for an Anchor.toml file.

    Returns:
        The location of the first Anchor.toml file found
    """
    search_dir = Path.cwd()
    root = Path(search_dir.root)

    while search_dir != root:
        attempt = search_dir / "Anchor.toml"
        if attempt.exists():
            return search_dir
        search_dir = search_dir.parent

    raise FileNotFoundError("Not in Anchor workspace.")


@app.callback()
def callback():
    """AnchorPy CLI."""


@app.command()
def shell():
    """Start IPython shell with AnchorPy workspace object initialized.

    Note that you should run `anchor localnet` before `anchorpy shell`.
    """
    path = _search_upwards_for_project_root()
    workspace = create_workspace(path)  # noqa: F841
    embed(
        colors="neutral",
        using="asyncio",
        banner2="Hint: type `workspace` to see the Anchor workspace object.\n",
    )


@app.command()
def init(
    program_name: str = typer.Argument(..., help="The name of the Anchor program.")
):
    """Create a basic Python test file for an Anchor program.

    This does not replace `anchor init`, but rather should be
    run after it.

    The test file will live at `tests/test_$PROGRAM_NAME.py`.
    """
    file_contents = INIT_TESTS.format(program_name)
    project_root = _search_upwards_for_project_root()
    file_path = project_root / "tests" / f"test_{program_name}.py"
    if file_path.exists():
        raise FileExistsError(file_path)
    file_path.write_text(file_contents)


@app.command()
def client_gen(
    idl: Path = typer.Argument(..., help="Anchor IDL file path"),
    out: Path = typer.Argument(..., help="Output directory."),
    program_id: Optional[str] = typer.Option(
        None, help="Optional program ID to be included in the code"
    ),
    pdas: bool = typer.Option(
        False, "--pdas", help="Auto-generate PDAs where possible."
    ),
):
    """Generate Python client code from the specified anchor IDL."""
    idl_obj = Idl.from_json(idl.read_text())
    if program_id is None:
        # Check for address in new format (top level)
        address_from_idl = idl_obj.address if hasattr(idl_obj, 'address') and idl_obj.address else None

        # Fall back to metadata.address (old format)
        if address_from_idl is None:
            idl_metadata = idl_obj.metadata
            address_from_idl = (
                idl_metadata.get("address") if isinstance(idl_metadata, dict) else None
            )

        if address_from_idl is None:
            typer.echo(
                "No program ID found in IDL. Use the --program-id "
                "option to set it manually."
            )
            raise typer.Exit(code=1)
        else:
            program_id_to_use = cast(str, address_from_idl)
    else:
        program_id_to_use = program_id

    typer.echo("generating package...")
    out.mkdir(exist_ok=True)
    (out / "__init__.py").touch()
    typer.echo("generating program_id.py...")
    gen_program_id(program_id_to_use, out)
    typer.echo("generating errors.py...")
    gen_errors(idl_obj, out)
    typer.echo("generating instructions...")
    gen_instructions(idl_obj, out, pdas)
    typer.echo("generating types...")
    gen_types(idl_obj, out)
    typer.echo("generating accounts...")
    gen_accounts(idl_obj, out)


if __name__ == "__main__":
    app()



import typing
from . import tile
from .tile import Tile, TileJSON
from . import game_state
from .game_state import GameStateKind, GameStateJSON
from . import sign
from .sign import SignKind, SignJSON



"""This module contains the invoke function."""
from asyncio import gather
from dataclasses import dataclass
from typing import NamedTuple, Optional, Union, cast

from solana.rpc.async_api import AsyncClient
from solana.rpc.commitment import Commitment, Confirmed, Finalized, Processed
from solana.rpc.core import RPCException
from solders.account import Account
from solders.account_decoder import UiAccountEncoding
from solders.commitment_config import CommitmentLevel
from solders.pubkey import Pubkey
from solders.rpc.config import RpcAccountInfoConfig
from solders.rpc.requests import GetMultipleAccounts, batch_to_json
from solders.rpc.responses import GetMultipleAccountsResp, RPCError, batch_from_json
from toolz import concat, partition_all

_GET_MULTIPLE_ACCOUNTS_LIMIT = 100
_MAX_ACCOUNT_SIZE = 10 * 1048576

_COMMITMENT_TO_SOLDERS = {
    Finalized: CommitmentLevel.Finalized,
    Confirmed: CommitmentLevel.Confirmed,
    Processed: CommitmentLevel.Processed,
}


class AccountInfo(NamedTuple):
    """Information describing an account.

    Attributes:
        executable: `True` if this account's data contains a loaded program.
        owner: Identifier of the program that owns the account.
        lamports: Number of lamports assigned to the account.
        data: Optional data assigned to the account.
        rent_epoch: Optional rent epoch info for for account.
    """

    executable: bool
    owner: Pubkey
    lamports: int
    data: bytes
    rent_epoch: Optional[int]

@dataclass
class _MultipleAccountsItem:
    pubkey: Pubkey
    account: Account


async def get_multiple_accounts(
    connection: AsyncClient,
    pubkeys: list[Pubkey],
    batch_size: int = 3,
    commitment: Optional[Commitment] = None,
) -> list[Optional[_MultipleAccountsItem]]:
    """Fetch multiple account infos through batched `getMultipleAccount` RPC requests.

    Args:
        connection: The `solana-py` client object.
        pubkeys: Pubkeys to fetch.
        batch_size: The number of `getMultipleAccount` objects to include in each
            HTTP request.
        commitment: Bank state to query.

    Returns:
        Account infos and pubkeys.
    """
    pubkeys_per_network_request = _GET_MULTIPLE_ACCOUNTS_LIMIT * batch_size
    chunks = partition_all(pubkeys_per_network_request, pubkeys)
    awaitables = [
        _get_multiple_accounts_core(connection, pubkeys_chunk, commitment)
        for pubkeys_chunk in chunks
    ]
    results = await gather(*awaitables, return_exceptions=False)
    return list(concat(results))


async def _get_multiple_accounts_core(
    connection: AsyncClient, pubkeys: list[Pubkey], commitment: Optional[Commitment]
) -> list[Optional[_MultipleAccountsItem]]:
    pubkey_batches = partition_all(_GET_MULTIPLE_ACCOUNTS_LIMIT, pubkeys)
    rpc_requests: list[GetMultipleAccounts] = []
    commitment_to_use = connection._commitment if commitment is None else commitment
    for pubkey_batch in pubkey_batches:
        rpc_req = GetMultipleAccounts(
            list(pubkey_batch),
            RpcAccountInfoConfig(
                encoding=UiAccountEncoding.Base64Zstd,
                commitment=_COMMITMENT_TO_SOLDERS[commitment_to_use],
            ),
        )
        rpc_requests.append(rpc_req)
    resp = await connection._provider.session.post(
        connection._provider.endpoint_uri,
        content=batch_to_json(rpc_requests),
        headers={"content-encoding": "gzip", "Content-type": "application/json"},
    )
    parsed = cast(
        list[Union[RPCError, GetMultipleAccountsResp]],
        batch_from_json(resp.text, [GetMultipleAccountsResp for _ in rpc_requests]),
    )
    result: list[Optional[_MultipleAccountsItem]] = []
    idx = 0
    for rpc_result in parsed:
        if not isinstance(rpc_result, GetMultipleAccountsResp):
            raise RPCException(f"Failed to get info about accounts: {rpc_result}")
        for account in rpc_result.value:
            if account is None:
                result.append(None)
            else:
                multiple_accounts_item = _MultipleAccountsItem(
                    pubkey=pubkeys[idx], account=account
                )
                result.append(multiple_accounts_item)
            idx += 1
    return result
